@@ -6,13 +6,14 @@ from contextlib import closing, contextmanager
 import re
 import six
 
+from _pytest.compat import dummy_context_manager
 from _pytest.config import create_terminal_writer
 import pytest
 import py
 
 
-DEFAULT_LOG_FORMAT = '%(filename)-25s %(lineno)4d %(levelname)-8s %(message)s'
-DEFAULT_LOG_DATE_FORMAT = '%H:%M:%S'
+DEFAULT_LOG_FORMAT = "%(filename)-25s %(lineno)4d %(levelname)-8s %(message)s"
+DEFAULT_LOG_DATE_FORMAT = "%H:%M:%S"
 
 
 class ColoredLevelFormatter(logging.Formatter):
@@ -21,19 +22,18 @@ class ColoredLevelFormatter(logging.Formatter):
     """
 
     LOGLEVEL_COLOROPTS = {
-        logging.CRITICAL: {'red'},
-        logging.ERROR: {'red', 'bold'},
-        logging.WARNING: {'yellow'},
-        logging.WARN: {'yellow'},
-        logging.INFO: {'green'},
-        logging.DEBUG: {'purple'},
+        logging.CRITICAL: {"red"},
+        logging.ERROR: {"red", "bold"},
+        logging.WARNING: {"yellow"},
+        logging.WARN: {"yellow"},
+        logging.INFO: {"green"},
+        logging.DEBUG: {"purple"},
         logging.NOTSET: set(),
     }
-    LEVELNAME_FMT_REGEX = re.compile(r'%\(levelname\)([+-]?\d*s)')
+    LEVELNAME_FMT_REGEX = re.compile(r"%\(levelname\)([+-]?\d*s)")
 
     def __init__(self, terminalwriter, *args, **kwargs):
-        super(ColoredLevelFormatter, self).__init__(
-            *args, **kwargs)
+        super(ColoredLevelFormatter, self).__init__(*args, **kwargs)
         if six.PY2:
             self._original_fmt = self._fmt
         else:
@@ -47,19 +47,20 @@ class ColoredLevelFormatter(logging.Formatter):
 
         for level, color_opts in self.LOGLEVEL_COLOROPTS.items():
             formatted_levelname = levelname_fmt % {
-                'levelname': logging.getLevelName(level)}
+                "levelname": logging.getLevelName(level)
+            }
 
             # add ANSI escape sequences around the formatted levelname
             color_kwargs = {name: True for name in color_opts}
             colorized_formatted_levelname = terminalwriter.markup(
-                formatted_levelname, **color_kwargs)
+                formatted_levelname, **color_kwargs
+            )
             self._level_to_fmt_mapping[level] = self.LEVELNAME_FMT_REGEX.sub(
-                colorized_formatted_levelname,
-                self._fmt)
+                colorized_formatted_levelname, self._fmt
+            )
 
     def format(self, record):
-        fmt = self._level_to_fmt_mapping.get(
-            record.levelno, self._original_fmt)
+        fmt = self._level_to_fmt_mapping.get(record.levelno, self._original_fmt)
         if six.PY2:
             self._fmt = fmt
         else:
@@ -78,61 +79,86 @@ def get_option_ini(config, *names):
 
 def pytest_addoption(parser):
     """Add options to control log capturing."""
-    group = parser.getgroup('logging')
+    group = parser.getgroup("logging")
 
     def add_option_ini(option, dest, default=None, type=None, **kwargs):
-        parser.addini(dest, default=default, type=type,
-                      help='default value for ' + option)
+        parser.addini(
+            dest, default=default, type=type, help="default value for " + option
+        )
         group.addoption(option, dest=dest, **kwargs)
 
     add_option_ini(
-        '--no-print-logs',
-        dest='log_print', action='store_const', const=False, default=True,
-        type='bool',
-        help='disable printing caught logs on failed tests.')
+        "--no-print-logs",
+        dest="log_print",
+        action="store_const",
+        const=False,
+        default=True,
+        type="bool",
+        help="disable printing caught logs on failed tests.",
+    )
     add_option_ini(
-        '--log-level',
-        dest='log_level', default=None,
-        help='logging level used by the logging module')
+        "--log-level",
+        dest="log_level",
+        default=None,
+        help="logging level used by the logging module",
+    )
     add_option_ini(
-        '--log-format',
-        dest='log_format', default=DEFAULT_LOG_FORMAT,
-        help='log format as used by the logging module.')
+        "--log-format",
+        dest="log_format",
+        default=DEFAULT_LOG_FORMAT,
+        help="log format as used by the logging module.",
+    )
     add_option_ini(
-        '--log-date-format',
-        dest='log_date_format', default=DEFAULT_LOG_DATE_FORMAT,
-        help='log date format as used by the logging module.')
+        "--log-date-format",
+        dest="log_date_format",
+        default=DEFAULT_LOG_DATE_FORMAT,
+        help="log date format as used by the logging module.",
+    )
     parser.addini(
-        'log_cli', default=False, type='bool',
-        help='enable log display during test run (also known as "live logging").')
+        "log_cli",
+        default=False,
+        type="bool",
+        help='enable log display during test run (also known as "live logging").',
+    )
     add_option_ini(
-        '--log-cli-level',
-        dest='log_cli_level', default=None,
-        help='cli logging level.')
+        "--log-cli-level", dest="log_cli_level", default=None, help="cli logging level."
+    )
     add_option_ini(
-        '--log-cli-format',
-        dest='log_cli_format', default=None,
-        help='log format as used by the logging module.')
+        "--log-cli-format",
+        dest="log_cli_format",
+        default=None,
+        help="log format as used by the logging module.",
+    )
     add_option_ini(
-        '--log-cli-date-format',
-        dest='log_cli_date_format', default=None,
-        help='log date format as used by the logging module.')
+        "--log-cli-date-format",
+        dest="log_cli_date_format",
+        default=None,
+        help="log date format as used by the logging module.",
+    )
     add_option_ini(
-        '--log-file',
-        dest='log_file', default=None,
-        help='path to a file when logging will be written to.')
+        "--log-file",
+        dest="log_file",
+        default=None,
+        help="path to a file when logging will be written to.",
+    )
     add_option_ini(
-        '--log-file-level',
-        dest='log_file_level', default=None,
-        help='log file logging level.')
+        "--log-file-level",
+        dest="log_file_level",
+        default=None,
+        help="log file logging level.",
+    )
     add_option_ini(
-        '--log-file-format',
-        dest='log_file_format', default=DEFAULT_LOG_FORMAT,
-        help='log format as used by the logging module.')
+        "--log-file-format",
+        dest="log_file_format",
+        default=DEFAULT_LOG_FORMAT,
+        help="log format as used by the logging module.",
+    )
     add_option_ini(
-        '--log-file-date-format',
-        dest='log_file_date_format', default=DEFAULT_LOG_DATE_FORMAT,
-        help='log date format as used by the logging module.')
+        "--log-file-date-format",
+        dest="log_file_date_format",
+        default=DEFAULT_LOG_DATE_FORMAT,
+        help="log date format as used by the logging module.",
+    )
 
 
 @contextmanager
@@ -153,7 +179,7 @@ def catching_logs(handler, formatter=None, level=None):
         root_logger.addHandler(handler)
     if level is not None:
         orig_level = root_logger.level
-        root_logger.setLevel(level)
+        root_logger.setLevel(min(orig_level, level))
     try:
         yield handler
     finally:
@@ -176,6 +202,10 @@ class LogCaptureHandler(logging.StreamHandler):
         self.records.append(record)
         logging.StreamHandler.emit(self, record)
 
+    def reset(self):
+        self.records = []
+        self.stream = py.io.TextIO()
+
 
 class LogCaptureFixture(object):
     """Provides access and control of log capturing."""
@@ -197,6 +227,9 @@ class LogCaptureFixture(object):
 
     @property
     def handler(self):
+        """
+        :rtype: LogCaptureHandler
+        """
         return self._item.catch_log_handler
 
     def get_records(self, when):
@@ -238,9 +271,25 @@ class LogCaptureFixture(object):
         """
         return [(r.name, r.levelno, r.getMessage()) for r in self.records]
 
+    @property
+    def messages(self):
+        """Returns a list of format-interpolated log messages.
+
+        Unlike 'records', which contains the format string and parameters for interpolation, log messages in this list
+        are all interpolated.
+        Unlike 'text', which contains the output from the handler, log messages in this list are unadorned with
+        levels, timestamps, etc, making exact comparisions more reliable.
+
+        Note that traceback or stack info (from :func:`logging.exception` or the `exc_info` or `stack_info` arguments
+        to the logging functions) is not included, as this is added by the formatter in the handler.
+
+        .. versionadded:: 3.7
+        """
+        return [r.getMessage() for r in self.records]
+
     def clear(self):
-        """Reset the list of log records."""
-        self.handler.records = []
+        """Reset the list of log records and the captured log text."""
+        self.handler.reset()
 
     def set_level(self, level, logger=None):
         """Sets the level for capturing of logs. The level will be restored to its previous value at the end of
@@ -282,9 +331,10 @@ def caplog(request):
 
     Captured logs are available through the following methods::
 
-    * caplog.text()          -> string containing formatted log output
-    * caplog.records()       -> list of logging.LogRecord instances
-    * caplog.record_tuples() -> list of (logger_name, level, message) tuples
+    * caplog.text            -> string containing formatted log output
+    * caplog.records         -> list of logging.LogRecord instances
+    * caplog.record_tuples   -> list of (logger_name, level, message) tuples
+    * caplog.clear()         -> clear captured records and formatted log output string
     """
     result = LogCaptureFixture(request.node)
     yield result
@@ -310,20 +360,14 @@ def get_actual_log_level(config, *setting_names):
     except ValueError:
         # Python logging does not recognise this as a logging level
         raise pytest.UsageError(
-            "'{0}' is not recognized as a logging level name for "
-            "'{1}'. Please consider passing the "
-            "logging level num instead.".format(
-                log_level,
-                setting_name))
+            "'{}' is not recognized as a logging level name for "
+            "'{}'. Please consider passing the "
+            "logging level num instead.".format(log_level, setting_name)
+        )
 
 
 def pytest_configure(config):
-    config.pluginmanager.register(LoggingPlugin(config), 'logging-plugin')
-
-
-@contextmanager
-def _dummy_context_manager():
-    yield
+    config.pluginmanager.register(LoggingPlugin(config), "logging-plugin")
 
 
 class LoggingPlugin(object):
@@ -339,25 +383,33 @@ class LoggingPlugin(object):
         self._config = config
 
         # enable verbose output automatically if live logging is enabled
-        if self._config.getini('log_cli') and not config.getoption('verbose'):
+        if self._log_cli_enabled() and not config.getoption("verbose"):
             # sanity check: terminal reporter should not have been loaded at this point
-            assert self._config.pluginmanager.get_plugin('terminalreporter') is None
+            assert self._config.pluginmanager.get_plugin("terminalreporter") is None
             config.option.verbose = 1
 
-        self.print_logs = get_option_ini(config, 'log_print')
-        self.formatter = logging.Formatter(get_option_ini(config, 'log_format'),
-                                           get_option_ini(config, 'log_date_format'))
-        self.log_level = get_actual_log_level(config, 'log_level')
+        self.print_logs = get_option_ini(config, "log_print")
+        self.formatter = logging.Formatter(
+            get_option_ini(config, "log_format"),
+            get_option_ini(config, "log_date_format"),
+        )
+        self.log_level = get_actual_log_level(config, "log_level")
 
-        log_file = get_option_ini(config, 'log_file')
+        log_file = get_option_ini(config, "log_file")
         if log_file:
-            self.log_file_level = get_actual_log_level(config, 'log_file_level')
+            self.log_file_level = get_actual_log_level(config, "log_file_level")
 
-            log_file_format = get_option_ini(config, 'log_file_format', 'log_format')
-            log_file_date_format = get_option_ini(config, 'log_file_date_format', 'log_date_format')
+            log_file_format = get_option_ini(config, "log_file_format", "log_format")
+            log_file_date_format = get_option_ini(
+                config, "log_file_date_format", "log_date_format"
+            )
             # Each pytest runtests session will write to a clean logfile
-            self.log_file_handler = logging.FileHandler(log_file, mode='w')
-            log_file_formatter = logging.Formatter(log_file_format, datefmt=log_file_date_format)
+            self.log_file_handler = logging.FileHandler(
+                log_file, mode="w", encoding="UTF-8"
+            )
+            log_file_formatter = logging.Formatter(
+                log_file_format, datefmt=log_file_date_format
+            )
             self.log_file_handler.setFormatter(log_file_formatter)
         else:
             self.log_file_handler = None
@@ -365,14 +417,28 @@ class LoggingPlugin(object):
         # initialized during pytest_runtestloop
         self.log_cli_handler = None
 
+    def _log_cli_enabled(self):
+        """Return True if log_cli should be considered enabled, either explicitly
+        or because --log-cli-level was given in the command-line.
+        """
+        return self._config.getoption(
+            "--log-cli-level"
+        ) is not None or self._config.getini("log_cli")
+
     @contextmanager
     def _runtest_for(self, item, when):
         """Implements the internals of pytest_runtest_xxx() hook."""
-        with catching_logs(LogCaptureHandler(),
-                           formatter=self.formatter, level=self.log_level) as log_handler:
+        with catching_logs(
+            LogCaptureHandler(), formatter=self.formatter, level=self.log_level
+        ) as log_handler:
             if self.log_cli_handler:
                 self.log_cli_handler.set_when(when)
-            if not hasattr(item, 'catch_log_handlers'):
+
+            if item is None:
+                yield  # run the test
+                return
+
+            if not hasattr(item, "catch_log_handlers"):
                 item.catch_log_handlers = {}
             item.catch_log_handlers[when] = log_handler
             item.catch_log_handler = log_handler
@@ -380,32 +446,40 @@ class LoggingPlugin(object):
                 yield  # run test
             finally:
                 del item.catch_log_handler
-                if when == 'teardown':
+                if when == "teardown":
                     del item.catch_log_handlers
 
             if self.print_logs:
                 # Add a captured log section to the report.
                 log = log_handler.stream.getvalue().strip()
-                item.add_report_section(when, 'log', log)
+                item.add_report_section(when, "log", log)
 
     @pytest.hookimpl(hookwrapper=True)
     def pytest_runtest_setup(self, item):
-        with self._runtest_for(item, 'setup'):
+        with self._runtest_for(item, "setup"):
             yield
 
     @pytest.hookimpl(hookwrapper=True)
     def pytest_runtest_call(self, item):
-        with self._runtest_for(item, 'call'):
+        with self._runtest_for(item, "call"):
             yield
 
     @pytest.hookimpl(hookwrapper=True)
     def pytest_runtest_teardown(self, item):
-        with self._runtest_for(item, 'teardown'):
+        with self._runtest_for(item, "teardown"):
             yield
 
+    @pytest.hookimpl(hookwrapper=True)
     def pytest_runtest_logstart(self):
         if self.log_cli_handler:
             self.log_cli_handler.reset()
+        with self._runtest_for(None, "start"):
+            yield
+
+    @pytest.hookimpl(hookwrapper=True)
+    def pytest_runtest_logfinish(self):
+        with self._runtest_for(None, "finish"):
+            yield
 
     @pytest.hookimpl(hookwrapper=True)
     def pytest_runtestloop(self, session):
@@ -414,8 +488,9 @@ class LoggingPlugin(object):
         with self.live_logs_context:
             if self.log_file_handler is not None:
                 with closing(self.log_file_handler):
-                    with catching_logs(self.log_file_handler,
-                                       level=self.log_file_level):
+                    with catching_logs(
+                        self.log_file_handler, level=self.log_file_level
+                    ):
                         yield  # run all the tests
             else:
                 yield  # run all the tests
@@ -425,22 +500,40 @@ class LoggingPlugin(object):
 
         This must be done right before starting the loop so we can access the terminal reporter plugin.
         """
-        terminal_reporter = self._config.pluginmanager.get_plugin('terminalreporter')
-        if self._config.getini('log_cli') and terminal_reporter is not None:
-            capture_manager = self._config.pluginmanager.get_plugin('capturemanager')
-            log_cli_handler = _LiveLoggingStreamHandler(terminal_reporter, capture_manager)
-            log_cli_format = get_option_ini(self._config, 'log_cli_format', 'log_format')
-            log_cli_date_format = get_option_ini(self._config, 'log_cli_date_format', 'log_date_format')
-            if self._config.option.color != 'no' and ColoredLevelFormatter.LEVELNAME_FMT_REGEX.search(log_cli_format):
-                log_cli_formatter = ColoredLevelFormatter(create_terminal_writer(self._config),
-                                                          log_cli_format, datefmt=log_cli_date_format)
+        terminal_reporter = self._config.pluginmanager.get_plugin("terminalreporter")
+        if self._log_cli_enabled() and terminal_reporter is not None:
+            capture_manager = self._config.pluginmanager.get_plugin("capturemanager")
+            log_cli_handler = _LiveLoggingStreamHandler(
+                terminal_reporter, capture_manager
+            )
+            log_cli_format = get_option_ini(
+                self._config, "log_cli_format", "log_format"
+            )
+            log_cli_date_format = get_option_ini(
+                self._config, "log_cli_date_format", "log_date_format"
+            )
+            if (
+                self._config.option.color != "no"
+                and ColoredLevelFormatter.LEVELNAME_FMT_REGEX.search(log_cli_format)
+            ):
+                log_cli_formatter = ColoredLevelFormatter(
+                    create_terminal_writer(self._config),
+                    log_cli_format,
+                    datefmt=log_cli_date_format,
+                )
             else:
-                log_cli_formatter = logging.Formatter(log_cli_format, datefmt=log_cli_date_format)
-            log_cli_level = get_actual_log_level(self._config, 'log_cli_level', 'log_level')
+                log_cli_formatter = logging.Formatter(
+                    log_cli_format, datefmt=log_cli_date_format
+                )
+            log_cli_level = get_actual_log_level(
+                self._config, "log_cli_level", "log_level"
+            )
             self.log_cli_handler = log_cli_handler
-            self.live_logs_context = catching_logs(log_cli_handler, formatter=log_cli_formatter, level=log_cli_level)
+            self.live_logs_context = catching_logs(
+                log_cli_handler, formatter=log_cli_formatter, level=log_cli_level
+            )
         else:
-            self.live_logs_context = _dummy_context_manager()
+            self.live_logs_context = dummy_context_manager()
 
 
 class _LiveLoggingStreamHandler(logging.StreamHandler):
@@ -461,6 +554,7 @@ class _LiveLoggingStreamHandler(logging.StreamHandler):
         self.capture_manager = capture_manager
         self.reset()
         self.set_when(None)
+        self._test_outcome_written = False
 
     def reset(self):
         """Reset the handler; should be called before the start of each test"""
@@ -470,18 +564,24 @@ class _LiveLoggingStreamHandler(logging.StreamHandler):
         """Prepares for the given test phase (setup/call/teardown)"""
         self._when = when
         self._section_name_shown = False
+        if when == "start":
+            self._test_outcome_written = False
 
     def emit(self, record):
-        if self.capture_manager is not None:
-            self.capture_manager.suspend_global_capture()
-        try:
-            if not self._first_record_emitted or self._when == 'teardown':
-                self.stream.write('\n')
+        ctx_manager = (
+            self.capture_manager.global_and_fixture_disabled()
+            if self.capture_manager
+            else dummy_context_manager()
+        )
+        with ctx_manager:
+            if not self._first_record_emitted:
+                self.stream.write("\n")
                 self._first_record_emitted = True
+            elif self._when in ("teardown", "finish"):
+                if not self._test_outcome_written:
+                    self._test_outcome_written = True
+                    self.stream.write("\n")
             if not self._section_name_shown and self._when:
-                self.stream.section('live log ' + self._when, sep='-', bold=True)
+                self.stream.section("live log " + self._when, sep="-", bold=True)
                 self._section_name_shown = True
             logging.StreamHandler.emit(self, record)
-        finally:
-            if self.capture_manager is not None:
-                self.capture_manager.resume_global_capture()
